package creim.analysis.audio;

import ddf.minim.AudioMetaData;
import ddf.minim.AudioPlayer;
import ddf.minim.AudioSample;
import ddf.minim.Minim;
import ddf.minim.analysis.FFT;
import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import javalib.util.Util;
import creim.util.Logger;
import org.apache.commons.math3.stat.StatUtils;



public class AudioFileAnalyzer {
	
	private Minim minim;
	private AudioSample song;
	private AudioMetaData meta;
	private AudioPlayer player;
	private String filename;
	private double[][] linearSpectrum;
	private double[][] logSpectrum;

	
	private PrintStream restore = System.out;
	private PrintStream shutup = new PrintStream(new OutputStream() {
		private StringBuilder sb = new StringBuilder();
		@Override
		public void write(int b) throws IOException {
			sb.append((char)b);
			if (b == '\n') {
				String s = sb.toString();
				if (!s.startsWith("===="))
					restore.write(s.getBytes());
			}
		}
	});
	
	
	public AudioFileAnalyzer() {
	}
	
	
	public AudioFileAnalyzer(String file) {
		this(new File(file));
	}
	
		
	public AudioFileAnalyzer(File file) {
		filename = file.getAbsolutePath();
		
		System.setOut(shutup);
		minim = new Minim(new MinimHelper());
		minim.debugOff();
		song = minim.loadSample(filename, 2048);
		meta = song.getMetaData();
		System.setOut(restore);
		
		System.out.printf("Song duration %d:%d\n", (int) (meta.length() / 1000 / 60), (int) (meta.length() / 1000 % 60));
	}
	
		
	
	public void play() {
		Logger.info("Playing song ... ");
		
		System.setOut(shutup);
		player = minim.loadFile(filename, 2048);
		player.play();
		System.setOut(restore);
		
		Util.slp(player.length());
		player.close();
	}
	
	
	public double[][] liearSpectra() {
		float[] leftChannel = song.getChannel(AudioSample.LEFT);
		float[] data = leftChannel;

		int fftSize = 2048;
		float[] fftSamples = new float[fftSize];
		FFT fft = new FFT(fftSize, song.sampleRate());

		int totalChunks = (data.length / fftSize) + 1;

		linearSpectrum = new double[totalChunks][fftSize / 2];

		for (int chunkIdx = 0; chunkIdx < totalChunks; ++chunkIdx) {
			int chunkStartIndex = chunkIdx * fftSize;
			int chunkSize = Math.min(data.length - chunkStartIndex, fftSize);

			System.arraycopy(data, chunkStartIndex, fftSamples, 0, chunkSize);

			if (chunkSize < fftSize) {
				java.util.Arrays.fill(fftSamples, chunkSize, fftSamples.length - 1, 0);
			}

			fft.forward(fftSamples);

			for (int i = 0; i < fftSize / 2; ++i) {
				linearSpectrum[chunkIdx][i] = fft.getBand(i);
			}
		}
		
		return linearSpectrum;
	}
	
	
	
	public double[][] logSpectrum(boolean db) {
		float[] leftChannel = song.getChannel(AudioSample.LEFT);
		float[] data = leftChannel;

		int fftSize = 2048;
		float[] fftSamples = new float[fftSize];
		FFT fft = new FFT(fftSize, song.sampleRate());
		fft.logAverages(200, 10);
		
		int totalChunks = (data.length / fftSize) + 1;
		logSpectrum = new double[totalChunks][fft.avgSize()];

		System.out.println("totalChunks:  "+totalChunks);
		System.out.println("FFT avg size: "+fft.avgSize());


		for (int chunkIdx = 0; chunkIdx < totalChunks; ++chunkIdx) {
			int chunkStartIndex = chunkIdx * fftSize;
			int chunkSize = Math.min(data.length - chunkStartIndex, fftSize);

			System.arraycopy(data, chunkStartIndex, fftSamples, 0, chunkSize);
			
			if (chunkSize < fftSize) {
				java.util.Arrays.fill(fftSamples, chunkSize, fftSamples.length - 1, 0);
			}

			fft.forward(fftSamples);

			for (int i = 0; i < fft.avgSize(); ++i) {
				if (db)
					logSpectrum[chunkIdx][i] = dB(Math.pow(Math.abs(fft.getAvg(i)), 2));
				else
					logSpectrum[chunkIdx][i] = fft.getAvg(i);
			}
		}

		return logSpectrum;
	}
	
	
	public double dB(double x) {
		if (x == 0) {
			return 0;
		} else {
			double d = 10 * Math.log10(x);
			return d < 0 ? 0 : d;
		}
	}
	
	
	public void close() {
		song.close();
	}
	
	
	public double[][] highs(double[][] fulllogdata) {
		double[][] lows = new double[fulllogdata.length][fulllogdata[0].length - 10];
		for (int i = 0; i < lows.length; i++) {
			for (int j = 0; j < lows[0].length; j++) {
				lows[i][j] = fulllogdata[i][10 + j];
			}
		}
		return lows;
	}
	
	
	public double[][] lows(double[][] fulllogdata) {
		double[][] lows = new double[fulllogdata.length][10];
		for (int i = 0; i < lows.length; i++) {
			for (int j = 0; j < lows[0].length; j++) {
				lows[i][j] = fulllogdata[i][j];
			}
		}
		return lows;
	}
	
	
	/**
	 * Energy variation.
	 * Smoothness of song from 0 to 1, where 0 is not smooth and 1 is very smooth.
	 * 
	 * Normalized average difference from timeslice to timeslice
	 * 
	 * @return 
	 */
	public double energySmoothness(double[][] data) {
		double[] avg = meanAmplitudesPerTime(data);
		double[] diffs = new double[avg.length-1];
		double x0 = avg[0];
		double maxdiff = 0;
		
		for (int x = 1; x < avg.length; x++) {
			double diff = Math.abs(avg[x] - x0);
			diffs[x-1] = diff;
			x0 = avg[x];
			
			if (diff > maxdiff) {
				maxdiff = diff;
			}
		}
		
		double sum = 0;
		for (int i = 0; i < diffs.length; i++) {
			sum += diffs[i];
		}
		double avgdiff = sum / diffs.length;

		return 1.0 - avgdiff / (1.0 - (avgdiff-maxdiff));
	}
	
	
	/**
	 * BROKEN!!!! NEEDS FIX
	 * Frequency variation.
	 * Smoothness of song from 0 to 1, where 0 is not smooth and 1 is very smooth.
	 * @return 
	 */
	@Deprecated
	public double frequencySmoothness(double[][] data) {
		//TODO: fix this broken shit
		double[] avg = meanFrequencies(data);
		double[] diffs = new double[avg.length-1];
		double x0 = avg[0];
		double maxdiff = 0;
		
		for (int x = 1; x < avg.length; x++) {
			double diff = Math.abs(avg[x] - x0);
			diffs[x-1] = diff;
			x0 = avg[x];
			
			if (diff > maxdiff) {
				maxdiff = diff;
			}
		}
		
		double sum = 0;
		for (int i = 0; i < diffs.length; i++) {
			sum += diffs[i];
		}
		double avgdiff = sum / diffs.length;

		return 1.0 - avgdiff / (1.0 - (avgdiff-maxdiff));
	}
	
	
	/**
	 * Average frequencies weighted by amplitude, over time.
	 * Detect tone direction, tones going higher vs lower
	 * @return 
	 */
	public double[] meanFrequencies(double[][] data) {
		int chunks = data.length;
		int freqs = data[0].length;
		double[] avgs = new double[chunks];
		for (int i = 0; i < chunks; i++) {
			double sum  = 0;
			double amps = 0;
			for (int j = 0; j < freqs; j++) {
				double d = Math.max(0, data[i][j]);
				sum += j * d;
				amps += d;
			}
			if (amps == 0)
				avgs[i] = 0;
			else
				avgs[i] = sum / amps;
		}
		return avgs;
	}
	
	
	
	/**
	 * Average amplitudes over all frequencies.
	 * @return 
	 */
	public double[] meanAmplitudesPerTime(double[][] data) {
		int chunks = data.length;
		int freqs = data[0].length;
		double[] res = new double[chunks];
		
		for (int c = 0; c < chunks; c++) {
			double ampsum = 0;
			for (int f = 0; f < freqs; f++) {
				ampsum += Math.max(0, data[c][f]);
			}
			double avgamp = ampsum / freqs;
			res[c] = avgamp;
		}
		
		return res;
	}
	
	
	
	/**
	 * Average amplitudes over time.
	 * @return 
	 */
	public double[] meanAmplitudesPerFrequency(double[][] data) {
		int chunks = data.length;
		int freqs = data[0].length;
		double[] res = new double[freqs];
		
		for (int f = 0; f < freqs; f++) {
			double ampsum = 0;
			for (int c = 0; c < chunks; c++) {
				ampsum += Math.max(0, data[c][f]);
			}
			double avgamp = ampsum / chunks;
			res[f] = avgamp;
		}
		
		return res;
	}
	
	
	
	/**
	 * 
	 * @return 
	 */
	public double[] stdFrequencies(double[][] data) {
		int chunks = data.length;
		int freqs = data[0].length;
		
		double[] stds = new double[chunks];
		double[] wms = meanFrequencies(data); // weighted mean over time. weighted by amplitude in a freqency
		
		for (int i = 0; i < chunks; i++) {
			int M = 0;
			double sum = 0;
			double wsum = 0;
			for (int x = 0; x < freqs; x++) {
				double w = Math.max(0, data[i][x]);
				if (w > 0)
					M++;
				sum += w * Math.pow(x - wms[i], 2);
				wsum += w;
				//System.out.printf("%d  %.2f\n", x, w);
			}
			//System.out.println("");
			if (M == 0) {
				stds[i] = 0;
			} 
			else {
				double std = Math.sqrt(sum / (((M-1) / (double)M) * wsum));
				stds[i] = std;
			}
		}
		
		return stds;
	}
	
	
	public double[] amplitudeVariancePerTime(double[][] data) {
		int chunks = data.length;
		int freqs = data[0].length;
		
		double[] vs = new double[chunks];
		for (int i = 0; i < chunks; i++) {
			double[] d = new double[freqs];
			System.arraycopy(data[i], 0, d, 0, freqs);
			
			double v = StatUtils.variance(d);
			vs[i] = v;
		}
		return vs;
	}
	
	
	public double[] amplitudeVariancePerFreqency(double[][] data) {
		int chunks = data.length;
		int freqs = data[0].length;
		
		double[] vs = new double[freqs];
		for (int i = 0; i < freqs; i++) {
			double[] d = new double[chunks];
			for (int j = 0; j < chunks; j++) {
				d[j] = data[j][i];
			}
			
			double v = StatUtils.variance(d);
			vs[i] = v;
		}
		return vs;
	}
	
	
	public double meanAmplitudeVariancePerTime(double[][] data) {
		return StatUtils.mean(amplitudeVariancePerTime(data));
	}
	
	
	public double meanAmplitudeVariancePerFreqency(double[][] data) {
		return StatUtils.mean(amplitudeVariancePerFreqency(data));
	}
	
}
