package creim.graphics.colors;

import java.awt.Color;
import javalib.util.MathUtil;
import javalib.util.RandomUtil;

public class ColorUtil {

	public static final float[][] colors7 = new float[][]{{0, 0, 0}, {0, 0, 1}, {0, 1, 0}, {1, 1, 0}, {1, 0, 0}, {0.75f, 0, 0}};
	public static final float[][] colors4 = new float[][]{{0, 0, 0}, {0, 0, 1}, {0, 1, 1}, {1, 1, 1}};
	public static final float[][] bluetored = new float[][]{{0, 0, 1}, {0, 1, 1}, {0, 1, 0}, {1, 1, 0}, {1, 0, 0}};

	public static int getRGBA(int r, int g, int b, int a) {
		int col = (a << 24) | (r << 16) | (g << 8) | b;
		return col;
	}

	public static int getRGBA(float r, float g, float b, float a) {
		return getRGBA((int) (255 * r), (int) (255 * g), (int) (255 * b), (int) (255 * a));
	}

	public static int HSBheatmap(double val, double min, double max) {
		float hue = (float) MathUtil.scale(val, min, max, 0.0, 1.0);
		return Color.getHSBColor(hue, 1.0f, 0.75f).getRGB();
	}

	public static Color monoChromeHeatmap(double val, double min, double max) {
		float heat = (float) MathUtil.scale(val, min, max, 0.0, 1.0);
		return new Color(heat, heat, heat);
	}

	public static Color monoChromeHeatmap(double val, double min, double max, Color from, Color to) {
		float value = (float) MathUtil.scale(val, min, max, 0.0, 1.0);
		int aR = from.getRed();
		int aG = from.getGreen();
		int aB = from.getBlue();
		int bR = to.getRed();
		int bG = to.getGreen();
		int bB = to.getBlue();

		int red = (int) ((bR - aR) * value + aR);      // Evaluated as -255*value + 255.
		int green = (int) ((bG - aG) * value + aG);      // Evaluates as 0.
		int blue = (int) ((bB - aB) * value + aB);      // Evaluates as 255*value + 0.
		return new Color(red, green, blue);
	}

	public static Color heatmap(double val, double min, double max) {
		return heatmap(val, min, max, colors7);
	}

	public static Color heatmap(double val, double min, double max, float[][] colors) {
		val = MathUtil.scale(val, min, max, 0.0, 1.0);

		int idx1;
		int idx2;
		double fractBetween = 0;

		if (val <= 0) {
			idx1 = idx2 = 0;
		} else if (val >= 1) {
			idx1 = idx2 = colors.length - 1;
		} else {
			val = val * (colors.length - 1);
			idx1 = (int) Math.floor(val);
			idx2 = idx1 + 1;
			fractBetween = val - (double) idx1;
		}

		float red = (float) ((colors[idx2][0] - colors[idx1][0]) * fractBetween + colors[idx1][0]);
		float green = (float) ((colors[idx2][1] - colors[idx1][1]) * fractBetween + colors[idx1][1]);
		float blue = (float) ((colors[idx2][2] - colors[idx1][2]) * fractBetween + colors[idx1][2]);

		return new Color(red, green, blue);
	}
	
	public static Color fromHEX(String colorStr) {
		return new Color(
			Integer.valueOf(colorStr.substring(1, 3), 16),
			Integer.valueOf(colorStr.substring(3, 5), 16),
			Integer.valueOf(colorStr.substring(5, 7), 16));
	}

	public static HSV RGB_TO_HSV(RGB in) {
		HSV out = new HSV();
		float min, max, delta;

		min = in.r < in.g ? in.r : in.g;
		min = min < in.b ? min : in.b;

		max = in.r > in.g ? in.r : in.g;
		max = max > in.b ? max : in.b;

		out.v = max;
		delta = max - min;
		if (delta < 0.00001) {
			out.s = 0;
			out.h = 0;
			return out;
		}
		if (max > 0.0) { // NOTE: if Max is == 0, this divide would cause a crash
			out.s = (delta / max);
		} else {
			// if max is 0, then r = g = b = 0              
			// s = 0, v is undefined
			out.s = 0.0f;
			out.h = Float.NaN;
			return out;
		}
		if (in.r >= max) {
			out.h = (in.g - in.b) / delta;
		} else if (in.g >= max) {
			out.h = 2.0f + (in.b - in.r) / delta;
		} else {
			out.h = 4.0f + (in.r - in.g) / delta;
		}

		out.h *= 60.0f;

		if (out.h < 0.0f) {
			out.h += 360.0f;
		}

		return out;
	}

	public static RGB HSV_TO_RGB(HSV in) {
		float hh, p, q, t, ff;
		int i;
		RGB out = new RGB();

		if (in.s <= 0.0) {
			out.r = in.v;
			out.g = in.v;
			out.b = in.v;
			return out;
		}
		hh = in.h;
		if (hh >= 360.0) {
			hh = 0.0f;
		}
		hh /= 60.0;
		i = (int) hh;
		ff = hh - i;
		p = in.v * (1.0f - in.s);
		q = in.v * (1.0f - (in.s * ff));
		t = in.v * (1.0f - (in.s * (1.0f - ff)));

		switch (i) {
			case 0:
				out.r = in.v;
				out.g = t;
				out.b = p;
				break;
			case 1:
				out.r = q;
				out.g = in.v;
				out.b = p;
				break;
			case 2:
				out.r = p;
				out.g = in.v;
				out.b = t;
				break;

			case 3:
				out.r = p;
				out.g = q;
				out.b = in.v;
				break;
			case 4:
				out.r = t;
				out.g = p;
				out.b = in.v;
				break;
			case 5:
			default:
				out.r = in.v;
				out.g = p;
				out.b = q;
				break;
		}
		return out;
	}

	public static RYB RGB_TO_RYB(RGB in) {
		float R = in.r;
		float G = in.g;
		float B = in.b;
		// remove the whiteness from the color
		float w = Math.min(Math.min(R, G), B);
		float r = R - w;
		float g = G - w;
		float b = B - w;

		float mg = Math.max(Math.max(r, g), b);

		// get the yellow out of the red and green
		float y = Math.min(r, g);
		r -= y;
		g -= y;

		// if this unfortunate conversion combines blue and green, then cut each in half to preserve the value's maximum range
		if (b != 0 && g != 0) {
			b /= 2.0;
			g /= 2.0;
		}

		// redistribute the remaining green
		y += g;
		b += g;

		// normalize to values
		float my = Math.max(r, Math.max(y, b));
		if (my != 0) {
			float n = mg / my;
			r *= n;
			y *= n;
			b *= n;
		}

		// add back in white
		r += w;
		y += w;
		b += w;

		RYB out = new RYB(r, y, b);

		return out;
	}

	@Deprecated
	public static RGB RYB_TO_RGB(RYB in) {
		float R = in.r;
		float Y = in.y;
		float B = in.b;
		// remove the whiteness from the color
		float w = Math.min(Math.min(R, Y), B);
		float r = R - w;
		float y = Y - w;
		float b = B - w;

		float my = Math.max(Math.max(r, y), b);

		// get the green out of the yellow and blue
		float g = Math.min(y, b);
		y -= g;
		b -= g;

		if (b != 0 && g != 0) {
			b *= 2.0;
			g *= 2.0;
		}

		// redistribute the remaining yellow
		r += y;
		g += y;

		// normalize to values
		float mg = Math.max(Math.max(r, g), b);
		if (mg != 0) {
			float n = my / mg;
			r *= n;
			g *= n;
			b *= n;
		}

		// add the white back in
		r += w;
		g += w;
		b += w;

		RGB out = new RGB(r, g, b);

		return out;
	}
	
	
	public static RYB mix(RYB c1, RYB c2) {
		RYB out = new RYB();
		
		out.r = c1.r + c2.r;
		out.y = c1.y + c2.y;
		out.b = c1.b + c2.b;
		
		float m = Math.max(out.r, Math.max(out.y, out.b));
		float f = 1 / m;
		out.r *= f;
		out.y *= f;
		out.b *= f;
		
		return out;
	}
	
	
	public static Color mixAvg(Color c1, Color c2) {
		int r = (c1.getRed() + c2.getRed()) / 2;
		int g = (c1.getGreen() + c2.getGreen()) / 2;
		int b = (c1.getBlue() + c2.getBlue()) / 2;
		int a = (c1.getAlpha() + c2.getAlpha()) / 2;
		
		return new Color(r, g, b, a);
	}
	
	
	public static Color mixMin(Color c1, Color c2) {
		int r = Math.min(255, c1.getRed() + c2.getRed());
		int g = Math.min(255, c1.getGreen() + c2.getGreen());
		int b = Math.min(255, c1.getBlue() + c2.getBlue());
		int a = Math.min(255, c1.getAlpha() + c2.getAlpha());
		
		return new Color(r, g, b, a);
	}
	
	
	public static int mixAvg(int c1, int c2) {
		int a1 = (c1 >> 24) & 0xff;
		int r1 = (c1 >> 16) & 0xff;
		int g1 = (c1 >>  8) & 0xff;
		int b1 = (c1      ) & 0xff;
		
		int a2 = (c2 >> 24) & 0xff;
		int r2 = (c2 >> 16) & 0xff;
		int g2 = (c2 >>  8) & 0xff;
		int b2 = (c2      ) & 0xff;
		
		int a  = (a1 + a2) / 2;
		int r  = (r1 + r2) / 2;
		int g  = (g1 + g2) / 2;
		int b  = (b1 + b2) / 2;
		
		int c  = ((a & 0xFF) << 24) |
                         ((r & 0xFF) << 16) |
                         ((g & 0xFF) <<  8) |
                         ((b & 0xFF)      );
		
		return c;
	}
	
	
	public static int mixMin(int c1, int c2) {
		int a1 = (c1 >> 24) & 0xff;
		int r1 = (c1 >> 16) & 0xff;
		int g1 = (c1 >>  8) & 0xff;
		int b1 = (c1      ) & 0xff;
		
		int a2 = (c2 >> 24) & 0xff;
		int r2 = (c2 >> 16) & 0xff;
		int g2 = (c2 >>  8) & 0xff;
		int b2 = (c2      ) & 0xff;
		
		int a  = Math.min(255, a1 + a2);
		int r  = Math.min(255, r1 + r2);
		int g  = Math.min(255, g1 + g2);
		int b  = Math.min(255, b1 + b2);
		
		int c  = ((a & 0xFF) << 24) |
                         ((r & 0xFF) << 16) |
                         ((g & 0xFF) <<  8) |
                         ((b & 0xFF)      );
		
		return c;
	}
	
	
	public static int mixNorm(int c1, int c2) {
		int a1 = (c1 >> 24) & 0xff;
		int r1 = (c1 >> 16) & 0xff;
		int g1 = (c1 >>  8) & 0xff;
		int b1 = (c1      ) & 0xff;
		
		int a2 = (c2 >> 24) & 0xff;
		int r2 = (c2 >> 16) & 0xff;
		int g2 = (c2 >>  8) & 0xff;
		int b2 = (c2      ) & 0xff;
		
		int a  = a1 + a2;
		int r  = r1 + r2;
		int g  = g1 + g2;
		int b  = b1 + b2;
		
		int max = Math.max(a, Math.max(r, Math.max(g, b)));
		double f = 255.0 / max;
		a = (int)(a * f);
		r = (int)(r * f);
		g = (int)(g * f);
		b = (int)(b * f);
		
		int c  = ((a & 0xFF) << 24) |
                         ((r & 0xFF) << 16) |
                         ((g & 0xFF) <<  8) |
                         ((b & 0xFF)      );
		
		return c;
	}
	
	
	public static Color randomColor() {
		return randomColor(false);
	}
	
	public static Color randomColor(boolean randalpha) {
		int r = RandomUtil.nextInt(256);
		int g = RandomUtil.nextInt(256);
		int b = RandomUtil.nextInt(256);
		int a = randalpha ? RandomUtil.nextInt(255) : 255;
		return new Color(r, g, b, a);
	}
	
	
	public static Color randomPastel() {
		return randomPastel(Color.WHITE);
	}
	
	
	public static Color randomPastel(Color mix) {
		return randomPastel(mix, false);
	}
	
	public static Color randomPastel(Color mix, boolean randalpha) {
		int r = RandomUtil.nextInt(256);
		int g = RandomUtil.nextInt(256);
		int b = RandomUtil.nextInt(256);
		int a = randalpha ? RandomUtil.nextInt(255) : 255;
		r = (r + mix.getRed()) / 2;
		g = (g + mix.getGreen()) / 2;
		b = (b + mix.getBlue()) / 2;
		return new Color(r, g, b, a);
	}
	
	
	public static Color scale(Color c, double factor) {
		int r = (int)Math.max(0, Math.min(255, Math.round(c.getRed() * factor)));
		int g = (int)Math.max(0, Math.min(255, Math.round(c.getGreen() * factor)));
		int b = (int)Math.max(0, Math.min(255, Math.round(c.getBlue() * factor)));
		int a = (int)Math.max(0, Math.min(255, Math.round(c.getAlpha() * factor)));
		
		return new Color(r, g, b, a);
	}
	
	
	public static int scale(int rgba, double f) {
		int a = (rgba >> 24) & 0xff;
		int r = (rgba >> 16) & 0xff;
		int g = (rgba >>  8) & 0xff;
		int b = (rgba      ) & 0xff;
		
		int na = (int)Math.max(0, Math.min(255, a * f));
		int nr = (int)Math.max(0, Math.min(255, r * f));
		int ng = (int)Math.max(0, Math.min(255, g * f));
		int nb = (int)Math.max(0, Math.min(255, b * f));
		
		int c  = ((na & 0xFF) << 24) |
                         ((nr & 0xFF) << 16) |
                         ((ng & 0xFF) <<  8) |
                         ((nb & 0xFF)      );
		
		return c;
	}
}
